# Copyright 2012 New Dream Network, LLC (DreamHost)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import context
from pecan import hooks


class ContextHook(hooks.PecanHook):
    """Moves the request env's neutron.context into the requests context."""
    priority = 95

    def before(self, state):
        ctx = (state.request.environ.get('neutron.context') or
               context.get_admin_context())
        state.request.context['neutron_context'] = ctx
