#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg

from neutron._i18n import _


L3GWMODE_OPTS = [
    cfg.BoolOpt('enable_snat_by_default', default=True,
                help=_('Define the default value of enable_snat if not '
                       'provided in external_gateway_info.'))
]


def register_db_l3_gwmode_opts(conf=cfg.CONF):
    conf.register_opts(L3GWMODE_OPTS)
